/*
 *
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2-eeprom.h"
#include "pvrusb2-hdw-low.h"
#include "pvrusb2-hdw-internal.h"
#include "pvrusb2-debug.h"

#define PVR_EEPROM_I2C_ADDR    0x50

#define trace_eeprom(...) pvr2_trace(PVR2_TRACE_EEPROM,__VA_ARGS__)

/*
 * Read and analyse data in the eeprom.  Use tveeprom to figure out
 * the packet structure, since this is another Hauppauge device and
 * internally it has a family resemblence to ivtv-type devices
 */

/* We seem to only be interested in the back half of the EEPROM */
#define EEPROM_SIZE 128
#define EEPROM_OFFS 128

int pvr2_eeprom_analyze(struct pvr2_hdw *hdw)
{
    u8 *eeprom;
    u8 offs;
    unsigned pcnt,tcnt;
    int ret;
    struct i2c_msg msg[2];

    memset(&hdw->tvdata,0,sizeof(hdw->tvdata));

    eeprom = kmalloc(EEPROM_SIZE,GFP_KERNEL);
    if (!eeprom) {
	return -ENOMEM;
    }

    msg[0].addr = PVR_EEPROM_I2C_ADDR;
    msg[0].flags = 0;
    msg[0].len = 1;
    msg[0].buf = &offs;
    msg[1].addr = PVR_EEPROM_I2C_ADDR;
    msg[1].flags = I2C_M_RD;

    /* We have to do the actual eeprom data fetch ourselves, because
       (1) we're only fetching part of the eeprom, and (2) if we were
       getting the whole thing our I2C driver can't grab it in one
       pass - which is what tveeprom is otherwise going to attempt */
    memset(eeprom,0,EEPROM_SIZE);
    for (tcnt = 0; tcnt < EEPROM_SIZE; tcnt += pcnt) {
	pcnt = 16;
	if (pcnt + tcnt > EEPROM_SIZE) pcnt = EEPROM_SIZE-tcnt;
	offs = tcnt + EEPROM_OFFS;
	msg[1].len = pcnt;
	msg[1].buf = eeprom+tcnt;
	if ((ret = i2c_transfer(&hdw->i2c_adap,
				msg,sizeof(msg)/sizeof(msg[0]))) != 2) {
	    trace_eeprom("eeprom fetch set offs err=%d",ret);
	    if (ret > 0) ret = -EIO;
	    kfree(eeprom);
	    return ret;
	}
    }

    tveeprom_hauppauge_analog(&hdw->tvdata,eeprom);

    trace_eeprom("eeprom results:");
    trace_eeprom("has_radio=%d",hdw->tvdata.has_radio);
    trace_eeprom("tuner_type=%d",hdw->tvdata.tuner_type);
    trace_eeprom("tuner_formats=0x%x",hdw->tvdata.tuner_formats);
    trace_eeprom("audio_processor=%d",hdw->tvdata.audio_processor);
    trace_eeprom("model=%d",hdw->tvdata.model);
    trace_eeprom("revision=%d",hdw->tvdata.revision);
    trace_eeprom("serial_number=%d",hdw->tvdata.serial_number);
    trace_eeprom("rev_str=%s",hdw->tvdata.rev_str);

    kfree(eeprom);

    return 0;
}


void pvr2_eeprom_set_default_standard(struct pvr2_hdw *hdw)
{
    int vstd_value;
    int vstd_found = 0;
    v4l2_std_id vs = (v4l2_std_id)hdw->tvdata.tuner_formats;

    if (vs & V4L2_STD_NTSC) {
	vstd_value = PVR2_CVAL_VIDEOSTANDARD_NTSC_M;
        vstd_found = !0;
    } else if (vs & V4L2_STD_PAL_BG) {
	vstd_value = PVR2_CVAL_VIDEOSTANDARD_PAL_BG;
        vstd_found = !0;
    } else if (vs & V4L2_STD_PAL_I) {
	vstd_value = PVR2_CVAL_VIDEOSTANDARD_PAL_I;
        vstd_found = !0;
    } else if (vs & V4L2_STD_PAL_DK) {
	vstd_value = PVR2_CVAL_VIDEOSTANDARD_PAL_DK;
        vstd_found = !0;
    } else if (vs & V4L2_STD_SECAM) {
	vstd_value = PVR2_CVAL_VIDEOSTANDARD_SECAM_L;
        vstd_found = !0;
    }

    if (!vstd_found) {
	trace_eeprom("eeprom unable to recognize a known video standard");
	return;
    }

    trace_eeprom("Setting video standard to %s (detected from eeprom)",
		 pvr2_hdw_get_ctl_value_name(hdw,
					     PVR2_CID_VIDEOSTANDARD,
					     vstd_value));
    pvr2_hdw_set_ctl_value_internal(hdw,PVR2_CID_VIDEOSTANDARD,vstd_value);
}


