/*
 *  $Id: pvrusb2-ioread.c,v 1.6 2005/04/09 04:45:22 isely Exp $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2-ioread.h"
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>

#define BUFFER_COUNT 32
#define BUFFER_SIZE PAGE_ALIGN(0x4000)

struct pvr2_ioread {
    struct pvr2_stream *stream;
    char *buffer_storage[BUFFER_COUNT];
    int resid_offs;
    int enabled;
    int stream_running;
    struct semaphore sem;
};

static int pvr2_ioread_init(struct pvr2_ioread *cp)
{
    unsigned int idx;

    cp->stream = 0;
    memset(cp,0,sizeof(*cp));
    init_MUTEX(&cp->sem);

    for (idx = 0; idx < BUFFER_COUNT; idx++) {
        cp->buffer_storage[idx] = kmalloc(BUFFER_SIZE,GFP_KERNEL);
        if (!(cp->buffer_storage[idx])) break;
    }

    if (idx < BUFFER_COUNT) {
        // An allocation appears to have failed
        for (idx = 0; idx < BUFFER_COUNT; idx++) {
            if (!(cp->buffer_storage[idx])) continue;
            kfree(cp->buffer_storage[idx]);
        }
        return -ENOMEM;
    }
    return 0;
}

static void pvr2_ioread_done(struct pvr2_ioread *cp)
{
    unsigned int idx;

    pvr2_ioread_setup(cp,0);
    for (idx = 0; idx < BUFFER_COUNT; idx++) {
        if (!(cp->buffer_storage[idx])) continue;
        kfree(cp->buffer_storage[idx]);
    }
}

struct pvr2_ioread *pvr2_ioread_create(void)
{
    struct pvr2_ioread *cp;
    cp = kmalloc(sizeof(*cp),GFP_KERNEL);
    if (pvr2_ioread_init(cp) < 0) {
        kfree(cp);
        return 0;
    }
    return cp;
}

void pvr2_ioread_destroy(struct pvr2_ioread *cp)
{
    pvr2_ioread_done(cp);
    kfree(cp);
}

static void pvr2_ioread_start(struct pvr2_ioread *cp)
{
    struct pvr2_buffer *bp;
    if (cp->enabled) return;
    if (!(cp->stream)) return;
    cp->resid_offs = 0;
    while ((bp = pvr2_stream_get_idle_buffer(cp->stream)) != 0) {
        pvr2_buffer_queue(bp);
    }
    cp->enabled = !0;
    cp->stream_running = 0;
}

static void pvr2_ioread_stop(struct pvr2_ioread *cp)
{
    if (!(cp->enabled)) return;
    pvr2_stream_kill(cp->stream);
    cp->resid_offs = 0;
    cp->enabled = 0;
    cp->stream_running = 0;
}

int pvr2_ioread_setup(struct pvr2_ioread *cp,struct pvr2_stream *sp)
{
    int ret;
    unsigned int idx;
    struct pvr2_buffer *bp;

    down(&cp->sem);
    if (cp->stream) {
        pvr2_ioread_stop(cp);
        pvr2_stream_kill(cp->stream);
        pvr2_stream_set_buffer_count(cp->stream,0);
        cp->stream = 0;
    }
    if (sp) {
        pvr2_stream_kill(sp);
        ret = pvr2_stream_set_buffer_count(sp,BUFFER_COUNT);
        if (ret < 0) return ret;
        for (idx = 0; idx < BUFFER_COUNT; idx++) {
            bp = pvr2_stream_get_buffer(sp,idx);
            pvr2_buffer_set_buffer(bp,cp->buffer_storage[idx],BUFFER_SIZE);
        }
        cp->stream = sp;
    }
    up(&cp->sem);

    return 0;
}

int pvr2_ioread_set_enabled(struct pvr2_ioread *cp,int fl)
{
    if ((!fl) == (!(cp->enabled))) return cp->enabled;

    down(&cp->sem);
    if (fl) {
        pvr2_ioread_start(cp);
    } else {
        pvr2_ioread_stop(cp);
    }
    up(&cp->sem);
    return !fl;
}

int pvr2_ioread_get_enabled(struct pvr2_ioread *cp)
{
    return cp->enabled != 0;
}

int pvr2_ioread_read(struct pvr2_ioread *cp,void __user *buf,unsigned int cnt)
{
    unsigned int copied_cnt;
    unsigned int bcnt;
    const char *src;
    int stat;
    struct pvr2_buffer *bp;
    if (!(cp->enabled)) {
        // Stream is not enabled; so this is an I/O error
        return -EIO;
    }

    if (!(cp->stream_running)) {
        if (pvr2_stream_get_ready_count(cp->stream) >= BUFFER_COUNT/2) {
            // Stream is enabled, and we've now pre-buffered up at least
            // half of our buffer space
            cp->stream_running = !0;
        } else {
            // Haven't buffered up enough yet; try again later
            return -EAGAIN;
        }
    }

    down(&cp->sem);

    // Suck data out of the buffers and copy to the user
    copied_cnt = 0;
    if (!buf) cnt = 0;
    while (cnt) {
        bp = pvr2_stream_get_ready_buffer(cp->stream);
        if (!bp) break; // Nothing ready; done

        bcnt = pvr2_buffer_get_count(bp);
        if (!bcnt) {
            // Nothing transferred.  Was there an error?
            stat = pvr2_buffer_get_status(bp);
            if (stat < 0) {
                // Streaming error; kill the stream
                pvr2_ioread_stop(cp);
                break;
            }
            src = 0;
        } else {
            // Calculate buffer offset and count
            src = cp->buffer_storage[pvr2_buffer_get_id(bp)];
            if (cp->resid_offs > bcnt) cp->resid_offs = bcnt;
            src += cp->resid_offs;
            bcnt -= cp->resid_offs;
        }
        if (bcnt > cnt) {
            // Can't read the entire buffer this time so remember how far
            // in we're going to get so we can resume it later
            cp->resid_offs += bcnt - cnt;
        } else {
            // We will be able to get the whole thing, so clear the
            // residual offset
            cp->resid_offs = 0;
        }
        if (bcnt) {
            if (copy_to_user(buf,src,bcnt)) {
                // User supplied a bad pointer?  Give up - this *will*
                // cause data to be lost.
                up(&cp->sem);
                return -EFAULT;
            }
            copied_cnt += bcnt;
        }
        if (cp->resid_offs) break; // If there's a residual offset, get out
        // Queue the buffer now that we've consumed it.  Note that we don't
        // bother with the return value (and doing so has serious
        // implications).  However the only possible error returned is if
        // we had passed in a null pointer, which can't happen if we've
        // gotten this far.
        pvr2_buffer_queue(bp);
    }

    up(&cp->sem);

    // If anything was copied, return that count
    if (copied_cnt) return copied_cnt;
    // Nothing copied; suggest to caller that attempt should be tried again
    // later
    return -EAGAIN;
}


/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
