/*
 *
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-buffer.h"
#include "pvrusb2-v4l2.h"
#include "pvrusb2-debug.h"

static ssize_t pvr_stream_read(struct file *, char *, size_t, loff_t *);
static unsigned int pvr_stream_poll(struct file *, poll_table *);
static int pvr_stream_mmap(struct file *, struct vm_area_struct *);

static struct file_operations pvr_fops = {
    .owner      = THIS_MODULE,
    .open       = pvr2_v4l2_open,
    .release    = pvr2_v4l2_release,
    .read       = pvr_stream_read,
    .ioctl      = pvr2_v4l2_ioctl,
    .llseek     = no_llseek,
    .poll       = pvr_stream_poll,
    .mmap       = pvr_stream_mmap,
};

static struct video_device pvr_video_template = {
    .owner      = THIS_MODULE,
    .type       = VID_TYPE_CAPTURE | VID_TYPE_TUNER,
    .type2      = (V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_VBI_CAPTURE |
                   V4L2_CAP_TUNER | V4L2_CAP_AUDIO |
                   V4L2_CAP_READWRITE | V4L2_CAP_STREAMING),
    .hardware   = VID_HARDWARE_PVRUSB2,
    .fops       = &pvr_fops,
};

int pvr_io_claim(struct pvr2_v4l2_stream *stream, enum pvr_io_method method,
                 int nbuff, struct pvr2_v4l2_fh *fh)
{
    int ret;
    struct pvr2_hdw *hdw = stream->v4lp->channel.hdw;

//    pvr_debug("pvr_io_claim");

    if (stream->io_method != _none) {
        if ((stream->io_method == method) && (stream->io_owner == fh))
            return 1;
        else
            return -EBUSY;
    }

    pvr2_hdw_set_streaming(hdw,0);

    if (nbuff)
        stream->buffer_count = min(nbuff, PVR_NBUFF);
    ret = pvr_buffer_alloc(stream);
    if (ret)
        return ret;

    if (method == _read){
        ret = pvr_buffer_submit(stream);
        if (ret) {
            pvr_buffer_free(stream);
            return ret;
        }
    }

    stream->io_method= method;
    stream->io_owner = fh;

    return 0;
}

int pvr_io_release(struct pvr2_v4l2_stream *stream, struct pvr2_v4l2_fh *fh)
{
//    pvr_debug("pvr_io_release");

    if (stream->io_owner != fh)
        return -EBUSY;

    switch(stream->io_method){
    case _read:
        pvr_buffer_unlink(stream);
        break;
    case _mmap:
        PVR_BUG_ON(stream->mem_state == _mapped);
        pvr_buffer_unlink(stream);
        break;
    case _uptr:
    case _none:
        return 0;
    }

    pvr_buffer_free(stream);
    stream->io_method = _none;
    stream->io_owner = NULL;

    return 0;
}

struct pvr2_v4l2_stream *pvr_stream_alloc(struct pvr2_v4l2 *v4lp, int endpoint,
                                          int v4l_type, int video_nr,
                                          enum pvr2_config config)
{
    struct pvr2_v4l2_stream *stream;
    struct usb_device *usbdev;

//    pvr_debug("pvr_stream_alloc");

    stream = kmalloc(sizeof(struct pvr2_v4l2_stream), GFP_KERNEL);
    if (stream == NULL)
        return NULL;

    memset(stream, 0, sizeof(struct pvr2_v4l2_stream));
    pvr2_trace(PVR2_TRACE_STRUCT,"Creating pvr_v4l2_stream id=0x%x",
               (unsigned int)stream);

    stream->v4lp = v4lp;
    init_MUTEX(&stream->stream_sem);
    init_waitqueue_head (&stream->wait_data);
    init_waitqueue_head (&stream->urbs_collected);
    spin_lock_init (&stream->ring_lock);

    stream->mem_state = _freed;
    stream->io_method = _none;
    stream->buffer_count = PVR_NBUFF;
    stream->endpoint = endpoint;
    stream->config = config;
    stream->v4l_type = v4l_type;

    stream->vdev = video_device_alloc();
    if (stream->vdev == NULL){
        err("alloc of vdev failed");
        goto err0;
    }

    memcpy(stream->vdev, &pvr_video_template, sizeof(struct video_device));
    usbdev = pvr2_hdw_get_dev(v4lp->channel.hdw);
    stream->vdev->dev = &usbdev->dev;
    stream->vdev->release = video_device_release;
    video_set_drvdata(stream->vdev, stream);

    if ((video_register_device(stream->vdev, v4l_type, video_nr) < 0) &&
        (video_register_device(stream->vdev, v4l_type, -1) < 0))
        goto err1;

    info("registered video device with minor %d", stream->vdev->minor & 0x1f);

    return stream;

 err1:
    err("unable to register video device with minor %d", video_nr);
    video_device_release(stream->vdev);
 err0:
    pvr2_trace(PVR2_TRACE_STRUCT,"Destroying pvr_v4l2_stream id=0x%x",
               (unsigned int)stream);
    kfree(stream);

    return NULL;
}

void pvr_stream_free(struct pvr2_v4l2_stream *stream)
{
//    pvr_debug("pvr_stream_free (%i)", stream->user);

    if (stream == NULL)
        return;

    PVR_BUG_ON(stream->io_method != _none);
    PVR_BUG_ON(stream->mem_state != _freed);

    video_unregister_device(stream->vdev);
    pvr2_trace(PVR2_TRACE_STRUCT,"Destroying pvr_v4l2_stream id=0x%x",
               (unsigned int)stream);
    kfree(stream);
}


/* Copy data to user space buffer 'dst', which is 'dsize' long. The 'src'
 * buffer contains 'ssize' bytes. 'pos' is the position in 'dst'.
 * Returns the number of bytes read in 'src' (can be different than number
 * of bytes written in 'dst').
 */

static int pvr_copy_to_user(u8 *dst, u8 *src, int dsize, int ssize,
                            enum pvr2_config type, int *pos)
{
    int i, n = 0;

//    pvr_debug("pvr_copy_to_user");

    switch(type) {
    case pvr2_config_mpeg:
    case pvr2_config_radio:
        n = min(dsize - *pos, ssize);
        (void) copy_to_user(dst + *pos, src, n);
        *pos += n;
        break;
    case pvr2_config_vbi:
        // FIXME : broken
        if (*pos == 0) {
            /* remove a 12 bytes header */
            n = min(dsize - 4, ssize - 12);
            if (n > 0) {
                copy_to_user(dst, src + 12, n);
                n += 12;
            } else {
                n = ssize;
                break;
            }
        } else {
            n = min(dsize - 4, ssize);
            copy_to_user(dst + *pos, src, n);
        }
        *pos += n;
        if (*pos == dsize - 4) {
            /* we must use a 4 bytes padding at the end, and swap data*/
            copy_to_user(dst + dsize - 4, dst, 4);
            *pos += 4;
            for (i = 0; i < dsize; i += 4)
                swab32s((u32 *)(dst + i));
        }
        break;
    default:
        BUG();
        break;
    }

    return n;
}

static ssize_t pvr_stream_read(struct file *file, char *buff, size_t count, loff_t *ppos)
{
    struct pvr2_v4l2_fh *fh = file->private_data;
    struct pvr2_v4l2_stream *stream = fh->stream;
    struct pvr_buffer *b;
    struct pvr2_v4l2 *vp = stream->v4lp;
    struct pvr2_hdw *hdw = vp->channel.hdw;
    int i, n;
    int ret = 0;

//    u8 *t = buff;
//    pvr_debug("pvr_stream_read : %i block : %i", count, file->f_flags & O_NONBLOCK);

    down(&stream->stream_sem); do {

        if (!pvr2_hdw_dev_ok(hdw)) {
            pvr_debug("device not ready");
            ret = -EIO;
            break;
        }

        ret =  pvr_io_claim(stream, _read, 0, fh);
        if (ret == 0) {
            /* first time we claim io. */
            pvr2_hdw_set_stream_type(hdw,stream->config);
            pvr2_hdw_set_streaming(hdw,!0);
            stream->last_buffer_timestamp = jiffies;
        }
        if (ret < 0) {
            break;
        }
        if (!count) {
            ret = 0;
            break;
        }

        if (file->f_flags & O_NONBLOCK) {
            count = min((int)count, (int)(PVR_NBUFF*PVR_VID_BUFFSIZE));
            /* Check if enough data is buffered. */
            ret = pvr_buffer_in(stream, count);
            if (ret){
                break;
            }
        }

        n = 0;
        ret = 0;
        while (n < count){
            int len;
            i = pvr_buffer_wait(stream,
                                ((file->f_flags & O_NONBLOCK) ?
                                 0 :
                                 2 * HZ));
            if (i < 0) {
                /* nothing for us. it can be -EIO or -EAGAIN */
                ret = i;
                break;
            }
            b = stream->rx_buffer[i % PVR_NBUFF];
            PVR_BUG_ON(b == NULL);
            len = pvr_copy_to_user(buff, b->data, count, b->size,
                                   stream->config, &n);
            b->size -= len;
            b->data += len;
            if (b->size == 0) {
                pvr_buffer_dq(stream);
                pvr_buffer_submit_one(stream, b->index);
            }
        }
        if (ret == 0) ret = n;

    } while(0); up(&stream->stream_sem);
//    pvr_debug("pvr_stream_read %i - %02x %02x %02x %02x %02x %02x ",
//        n, t[0],t[1],t[2],t[3],t[4],t[5] );

//    pvr_debug("read %d byte(s)",n);
    return ret;
}

static int pvr_stream_mmap(struct file *file, struct vm_area_struct *vma)
{
    struct pvr2_v4l2_fh *fh = file->private_data;
    struct pvr2_v4l2_stream *stream = fh->stream;
    int ret;

//    pvr_debug("pvr_stream_mmap");

    down(&stream->stream_sem); do {
        ret = pvr_map_buffer(stream, vma);
    } while (0); up(&stream->stream_sem);

    return ret;
}

/* FIXME : Not sure if this one is correct ! */

static unsigned int pvr_stream_poll(struct file *file, poll_table *wait)
{
    struct pvr2_v4l2_fh *fh = file->private_data;
    struct pvr2_v4l2_stream *stream = fh->stream;
    unsigned int mask = 0;

//    pvr_debug("pvr_stream_poll: %i ", stream->io_method);

    if (stream->io_method == _none)
        return POLLERR;

    /* add stream's waitq to poll list */
    poll_wait(file, &stream->wait_data, wait);

    if (stream->rx1 != stream->rx2)
        mask |= POLLIN | POLLRDNORM;

    return mask;
}

/* Streaming API */

#define PVR_COMPAT(t1, t2)    (\
    ((t1 == VFL_TYPE_GRABBER) && (t2 == V4L2_BUF_TYPE_VIDEO_CAPTURE))    \
    || ((t1 == VFL_TYPE_VBI) && (t2 == V4L2_BUF_TYPE_VBI_CAPTURE))  \
)

int pvr_stream_qbuf(struct pvr2_v4l2_stream *stream, struct v4l2_buffer *vb)
{
    int ret = -EINVAL;

//    pvr_debug("pvr_stream_qbuf");

    if (!PVR_COMPAT(stream->v4l_type, vb->type) || vb->index < 0
        || vb->index >= stream->buffer_count)
        return -EINVAL;

    if (vb->memory == V4L2_MEMORY_MMAP) {
        down(&stream->stream_sem); do {
            ret = pvr_buffer_submit_one(stream, vb->index);
        } while (0); up(&stream->stream_sem);

        if (ret == 0){
            vb->flags |= V4L2_BUF_FLAG_MAPPED | V4L2_BUF_FLAG_QUEUED;
            vb->flags &= ~V4L2_BUF_FLAG_DONE;
        } else
            ret = -EINVAL;
    } else if (vb->memory == V4L2_MEMORY_USERPTR) {
        down(&stream->stream_sem); do {
            stream->userptr[stream->uptr2 % PVR_NBUFF].ptr = vb->m.userptr;
            stream->userptr[stream->uptr2 % PVR_NBUFF].length = vb->length;
            stream->uptr2++;
        } while (0); up(&stream->stream_sem);
        vb->flags |= V4L2_BUF_FLAG_QUEUED;
        vb->flags &= ~V4L2_BUF_FLAG_DONE & ~V4L2_BUF_FLAG_MAPPED;
        ret = 0;
    }

    return ret;
}

int pvr_stream_querybuf(struct pvr2_v4l2_stream *stream, struct v4l2_buffer *vb)
{

//    pvr_debug("pvr_stream_querybuf");

    if (!PVR_COMPAT(stream->v4l_type, vb->type)
        || vb->index < 0 || vb->index >= stream->buffer_count)
        return -EINVAL;

    vb->memory = V4L2_MEMORY_MMAP;
    vb->flags = 0;

    down(&stream->stream_sem); do {

        if (stream->buffer[vb->index].map)
            vb->flags |= V4L2_BUF_FLAG_MAPPED;

        if (stream->buffer[vb->index].submitted)
            vb->flags |= V4L2_BUF_FLAG_QUEUED;
        else if (stream->buffer[vb->index].data)
            vb->flags |= V4L2_BUF_FLAG_DONE;

        vb->length = PVR_VID_BUFFSIZE;
        vb->m.offset = vb->index * PVR_VID_BUFFSIZE;

    } while (0); up(&stream->stream_sem);

    return 0;
}

int pvr_stream_dqbuf(struct pvr2_v4l2_stream *stream,
                     struct v4l2_buffer *vb, int nonblock)
{
    int i, ret = 0;

//    pvr_debug("pvr_stream_dqbuf");

    if (!PVR_COMPAT(stream->v4l_type, vb->type))
//        || vb->memory != V4L2_MEMORY_MMAP)
        return -EINVAL;

    down(&stream->stream_sem); do {

        if (vb->memory == V4L2_MEMORY_MMAP) {
            i = pvr_buffer_wait(stream, nonblock ? 0 : 2 * HZ);
            if (i >= 0){
                vb->index = stream->rx_buffer[i]->index;
                vb->bytesused = stream->rx_buffer[i]->size;
                vb->flags = V4L2_BUF_FLAG_DONE;
                vb->flags |= (stream->rx_buffer[i]->map ? V4L2_BUF_FLAG_MAPPED : 0);
                vb->field = V4L2_FIELD_NONE;
                vb->sequence = 1;
                vb->m.offset = vb->index * PVR_VID_BUFFSIZE;
                vb->length = PVR_VID_BUFFSIZE;

                /* dequeue buffer */
                stream->rx_buffer[i]->data = NULL;
                stream->rx1++;
            } else
                ret = i;
        } else if (vb->memory == V4L2_MEMORY_USERPTR) {
            int count = stream->userptr[stream->uptr2 % PVR_NBUFF].length;
            u8* buff = (u8*) stream->userptr[stream->uptr2 % PVR_NBUFF].ptr;
            int n;

            /*  */
            if (nonblock) {
                count = min(count, (int)(PVR_NBUFF*PVR_VID_BUFFSIZE));
                /* Check if enough data is buffered. */
                ret = pvr_buffer_in(stream, count);
                if (ret) {
                    break;
                }
            }

            n = 0;
            ret = 0;
            while (n < count){
                int len;
                struct pvr_buffer *b;
                i = pvr_buffer_wait(stream, nonblock ? 0 : 2 * HZ);
                if (i < 0) {
                    /* nothing for us. it can be -EIO or -EAGAIN */
                    ret = i;
                    break;
                }
                b = stream->rx_buffer[i % PVR_NBUFF];
                PVR_BUG_ON(b == NULL);
                len = pvr_copy_to_user(buff, b->data, count, b->size,
                                       stream->config, &n);
                b->size -= len;
                b->data += len;
                if (b->size == 0) {
                    pvr_buffer_dq(stream);
                    pvr_buffer_submit_one(stream, b->index);
                }
            }
            if (ret) break;
            vb->length = n;
            vb->bytesused = n;
            vb->m.userptr = (unsigned long) buff;
            vb->flags |= V4L2_BUF_FLAG_DONE;
            vb->flags &= ~V4L2_BUF_FLAG_QUEUED & ~V4L2_BUF_FLAG_MAPPED;
            ret = (stream->uptr2++) % PVR_NBUFF;
        } else
            ret = -EINVAL;

    } while (0); up(&stream->stream_sem);

    return ret;
}

int pvr_stream_reqbufs(struct pvr2_v4l2_stream *stream, struct pvr2_v4l2_fh *fh,
                       struct v4l2_requestbuffers *vr)
{
    int ret = 0, count = PVR_NBUFF;
    enum pvr_io_method io;

//    pvr_debug("pvr_stream_reqbufs");

    if (!PVR_COMPAT(stream->v4l_type, vr->type))
        return -EINVAL;

    if (vr->memory == V4L2_MEMORY_USERPTR)
        io = _uptr;
    else if ((vr->memory == V4L2_MEMORY_MMAP) &&
             (vr->type != V4L2_BUF_TYPE_VBI_CAPTURE)) {
        io = _mmap;
        if (vr->count < PVR_NBUFF)
            count = vr->count;
    } else
        return -EINVAL;

    down(&stream->stream_sem); do {
        if (count) {
            ret = pvr_io_claim(stream, io, count, fh);
            if ((ret == 1) && (count != stream->buffer_count)) {
                err("Retrying pvr_io_claim?");
                pvr_io_release(stream, fh);
                ret = pvr_io_claim(stream, io, count, fh);
            }
        } else {
            err("count == 0 when requesting buffers?");
            ret = pvr_io_release(stream, fh);
        }
    } while (0); up(&stream->stream_sem);

    return ret;
}

int pvr_stream_streamon(struct pvr2_v4l2_stream *stream, struct pvr2_v4l2_fh *fh)
{
    int ret, retry = 1;
    struct pvr2_v4l2 *vp = stream->v4lp;
    struct pvr2_hdw *hdw = vp->channel.hdw;

//    pvr_debug("pvr_stream_streamon");

    if (stream->mem_state < _allocated)
        return -EINVAL;
 re:
    ret = pvr2_hdw_set_stream_type(hdw,stream->config);
    if (!ret) pvr2_hdw_set_streaming(hdw,!0);
    if (ret < 0) return ret;

    down(&stream->stream_sem); do {
        ret =  pvr_buffer_wait(stream, 2 * HZ);
    } while (0); up(&stream->stream_sem);

    if ((ret < 0) && retry){
        /* timeout. retry. */
        retry = 0;
        goto re;
    }

    return ret;
}

int pvr_stream_streamoff(struct pvr2_v4l2_stream *stream, struct pvr2_v4l2_fh *fh)
{
    struct pvr2_v4l2 *vp = stream->v4lp;
    struct pvr2_hdw *hdw = vp->channel.hdw;
//    pvr_debug("pvr_stream_streamoff");

    pvr2_hdw_set_streaming(hdw,0);
    pvr2_hdw_set_stream_type(hdw,pvr2_config_empty);

    down(&stream->stream_sem); do {
        pvr_buffer_unlink(stream);
    } while (0); up(&stream->stream_sem);

    return 0;
}

int pvr_stream_cgmbuf(struct pvr2_v4l2_stream *stream, struct pvr2_v4l2_fh *fh,
                      struct video_mbuf *vmb)
{
    int ret, i, n = min(PVR_NBUFF, VIDEO_MAX_FRAME);

//    pvr_debug("pvr_stream_cgmbuf");

    down(&stream->stream_sem); do {
        ret = pvr_io_claim(stream, _mmap, n , fh);
    } while (0); up(&stream->stream_sem);
    if (ret)
        return ret;
    vmb->size = stream->buffer_count * PVR_VID_BUFFSIZE;
    vmb->frames = stream->buffer_count;
    for (i = 0 ; i < stream->buffer_count ; i++)
        vmb->offsets[i] = PVR_VID_BUFFSIZE * i;

    return 0;
}

/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
